from typing import Callable, List, TypeVar

T = TypeVar('T')


def parse_ranking_with_ties(ranking: str, converter: Callable[[str], T]) -> List[List[T]]:
    ranking = ranking.strip()

    # to manage the "syn" datasets of java rank-n-ties
    if ranking[-1] == ":":
        ranking = ranking[:-1]
    ret = []
    st = ranking.find('[', ranking.find('[') + 1)
    en = ranking.find(']')
    ranking_end = ranking.rfind(']')
    old_en = en
    if ranking[ranking_end + 1:] != "":
        raise ValueError("remaining chars at the end: '%s'" % ranking[ranking_end + 1:])
    while st != -1 and en != -1:
        bucket = []
        for s in ranking[st + 1:  en].split(","):
            elt_str = s.strip()
            if elt_str == "":
                raise ValueError("Empty element in `%s` between chars %i and %i" % (ranking[st + 1:  en], st + 1, en))
            bucket.append(converter(elt_str))
        ret.append(bucket)
        st = ranking.find('[', en + 1, ranking_end)
        old_en = en
        en = ranking.find(']', max(en + 1, st + 1), ranking_end)
    if st != en:
        if st == -1:
            raise ValueError("missing open bucket")
        if en == -1:
            raise ValueError("missing closing bucket")
    if ranking[old_en + 1:ranking_end] != "":
        raise ValueError("misplaced chars: '%s'" % ranking[old_en + 1:ranking_end])
    return ret


def parse_ranking_with_ties_of_str(ranking: str) -> List[List[str]]:
    return parse_ranking_with_ties(
        ranking=ranking,
        converter=lambda x: x
    )


def parse_ranking_with_ties_of_int(ranking: str) -> List[List[int]]:
    return parse_ranking_with_ties(
        ranking=ranking,
        converter=lambda x: int(x)
    )


def get_rankings_from_file(file: str) -> List[List[List[int or str]]]:
    rankings = []
    rankings_int = []
    file_rankings = open(file, "r")
    only_ints = True
    # to manage the "step" datasets of java rank-n-ties
    ignore_lines = ["%"]
    lignes = file_rankings.read().replace("\\\n", "")
    for ligne in lignes.split("\n"):
        if len(ligne) > 2 and ligne[0] not in ignore_lines:
            ranking = parse_ranking_with_ties_of_str(ligne)
            if only_ints:
                if is_ranking_of_int(ranking):
                    rankings_int.append(parse_ranking_with_ties_of_int(ligne))
                else:
                    only_ints = False
            rankings.append(ranking)
    file_rankings.close()
    if only_ints:
        return rankings_int
    return rankings


def dump_ranking_with_ties_to_str(ranking: List[List[int or str]]) -> str:
    return '[' + ','.join(['[' + ','.join([str(e) for e in b]) + ']' for b in ranking]) + ']'


def is_ranking_of_int(ranking: List[List[str]]) -> bool:
    for bucket in ranking:
        for elem in bucket:
            if not elem.isdigit():
                return False
    return True


def rankings_from_str_to_int(ranking: List[List[int or str]])->List[List[int]]:
    res = []
    for bucket in ranking:
        res.append([int(elem) for elem in bucket])
    return res



from typing import List, Dict, Tuple, Set
from corankco.algorithms.median_ranking import MedianRanking
from corankco.dataset import Dataset
from corankco.scoringscheme import ScoringScheme
from corankco.consensus import Consensus, ConsensusFeature
from numpy import ndarray, array, shape, zeros, count_nonzero, vdot, asarray
from operator import itemgetter
import pulp
from igraph import Graph


class ExactAlgorithmGeneric(MedianRanking):
    def __init__(self, limit_time_sec=0, scoring_scheme=None):
        if limit_time_sec > 0:
            self.__limit_time_sec = limit_time_sec
        else:
            self.__limit_time_sec = 0
        self.__scoring_scheme = scoring_scheme

    def compute_consensus_rankings(
            self,
            dataset: Dataset,
            scoring_scheme: ScoringScheme,
            return_at_most_one_ranking: bool=False,
            bench_mode: bool = False
    ) -> Consensus:
        """
        :param dataset: A dataset containing the rankings to aggregate
        :type dataset: Dataset (class Dataset in package 'datasets')
        :param scoring_scheme: The penalty vectors to consider
        :type scoring_scheme: ScoringScheme (class ScoringScheme in package 'distances')
        :param return_at_most_one_ranking: the algorithm should not return more than one ranking
        :type return_at_most_one_ranking: bool
        :param bench_mode: is bench mode activated. If False, the algorithm may return more information
        :type bench_mode: bool
        :return one or more rankings if the underlying algorithm can find several equivalent consensus rankings
        If the algorithm is not able to provide multiple consensus, or if return_at_most_one_ranking is True then, it
        should return a list made of the only / the first consensus found.
        In all scenario, the algorithm returns a list of consensus rankings
        :raise ScoringSchemeNotHandledException when the algorithm cannot compute the consensus because the
        implementation of the algorithm does not fit with the scoring scheme
        """
        rankings = dataset.rankings
        elem_id = {}
        id_elements = {}
        id_elem = 0
        for ranking in rankings:
            for bucket in ranking:
                for element in bucket:
                    if element not in elem_id:
                        elem_id[element] = id_elem
                        id_elements[id_elem] = element
                        id_elem += 1
        nb_elem = len(elem_id)

        positions = ExactAlgorithmGeneric.__positions(rankings, elem_id)

        sc = asarray(scoring_scheme.penalty_vectors_str)

        graph, mat_score, ties_must_be_checked = self.__graph_of_elements(positions, sc)

        my_values = []
        my_vars = []

        h_vars = {}
        cpt = 0

        for i in range(nb_elem):
            for j in range(nb_elem):
                if not i == j:
                    name_var = "x_%s_%s" % (i, j)
                    my_values.append(mat_score[i][j][0])
                    my_vars.append(pulp.LpVariable(name_var, 0, 1, cat="Binary"))
                    h_vars[name_var] = cpt
                    cpt += 1
                    if i < j:
                        name_var = "t_%s_%s" % (i, j)
                        my_values.append(mat_score[i][j][2])
                        my_vars.append(pulp.LpVariable(name_var, 0, 1, cat="Binary"))
                        h_vars[name_var] = cpt
                        cpt += 1

        prob = pulp.LpProblem("myProblem", pulp.LpMinimize)

        # add the binary order constraints
        for i in range(0, nb_elem - 1):
            for j in range(i + 1, nb_elem):
                if not i == j:
                    prob += my_vars[h_vars["x_%s_%s" % (i, j)]] \
                            + my_vars[h_vars["x_%s_%s" % (j, i)]] \
                            + my_vars[h_vars["t_%s_%s" % (i, j)]] == 1

        # add the transitivity constraints
        for i in range(0, nb_elem):
            for j in range(nb_elem):
                if j != i:
                    i_bef_j = "x_%s_%s" % (i, j)
                    if i < j:
                        i_tie_j = "t_%s_%s" % (i, j)
                    else:
                        i_tie_j = "t_%s_%s" % (j, i)
                    for k in range(nb_elem):
                        if k != i and k != j:
                            j_bef_k = "x_%s_%s" % (j, k)
                            i_bef_k = "x_%s_%s" % (i, k)
                            if j < k:
                                j_tie_k = "t_%s_%s" % (j, k)
                            else:
                                j_tie_k = "t_%s_%s" % (k, j)

                            if i < k:
                                i_tie_k = "t_%s_%s" % (i, k)
                            else:
                                i_tie_k = "t_%s_%s" % (k, i)

                            prob += my_vars[h_vars[i_bef_j]] +\
                                my_vars[h_vars[j_bef_k]] \
                                + my_vars[h_vars[j_tie_k]] \
                                - my_vars[h_vars[i_bef_k]] <= 1

                            prob += my_vars[h_vars[i_bef_j]] + \
                                my_vars[h_vars[i_tie_j]] \
                                + my_vars[h_vars[j_bef_k]] - my_vars[h_vars[i_bef_k]] <= 1

                            prob += 2 * my_vars[h_vars[i_tie_j]] \
                                + 2 * my_vars[h_vars[j_tie_k]] \
                                - my_vars[h_vars[i_tie_k]] <= 3

        # if tie is not the single best choice for any pair x,y, then there is a permutation median
        if not ties_must_be_checked:
            for i in range(0, nb_elem - 1):
                for j in range(i + 1, nb_elem):
                    if not i == j:
                        prob += my_vars[h_vars["t_%s_%s" % (i, j)]] == 0

        cfc = graph.components()

        for i in range(len(cfc)):
            group_i = cfc[i]
            for j in range(i+1, len(cfc)):
                for elem_i in group_i:
                    for elem_j in cfc[j]:
                        prob += my_vars[h_vars["x_%s_%s" % (elem_i, elem_j)]] == 1
                        prob += my_vars[h_vars["x_%s_%s" % (elem_j, elem_i)]] == 0
                        if elem_i < elem_j:
                            prob += my_vars[h_vars["t_%s_%s" % (elem_i, elem_j)]] == 0
                        else:
                            prob += my_vars[h_vars["t_%s_%s" % (elem_j, elem_i)]] == 0

        # objective function
        prob += pulp.lpSum(my_vars[cpt] * my_values[cpt] for cpt in range(len(my_vars)))

        try:
            prob.solve(pulp.CPLEX(msg=False))
        except:
            prob.solve(pulp.PULP_CBC_CMD(msg=False))

        h_def = {i: 0 for i in range(nb_elem)}

        for var in my_vars:
            if abs(var.value() - 1) < 0.01 and var.name[0] == "x":
                h_def[int(var.name.split("_")[2])] += 1

        ranking = []
        current_nb_def = 0
        bucket = []

        for elem, nb_defeats in (sorted(h_def.items(), key=itemgetter(1))):
            if nb_defeats == current_nb_def:
                bucket.append(id_elements[elem])
            else:
                ranking.append(bucket)
                bucket = [id_elements[elem]]
                current_nb_def = nb_defeats
        ranking.append(bucket)
        return Consensus(consensus_rankings=[ranking],
                         dataset=dataset,
                         scoring_scheme=scoring_scheme,
                         att={ConsensusFeature.IsNecessarilyOptimal: True,
                              ConsensusFeature.KemenyScore: prob.objective.value(),
                              ConsensusFeature.AssociatedAlgorithm: self.get_full_name()
                              })

    @staticmethod
    def __graph_of_elements(positions: ndarray, matrix_scoring_scheme: ndarray) -> Tuple[Graph, ndarray, bool]:
        graph_of_elements = Graph(directed=True)
        cost_before = matrix_scoring_scheme[0]
        cost_tied = matrix_scoring_scheme[1]
        cost_after = array([cost_before[1], cost_before[0], cost_before[2], cost_before[4], cost_before[3],
                            cost_before[5]])
        n = shape(positions)[0]
        m = shape(positions)[1]
        ties_must_be_checked = False

        for i in range(n):
            graph_of_elements.add_vertex(name=str(i))

        matrix = zeros((n, n, 3))
        edges = []
        for e1 in range(n):
            mem = positions[e1]
            d = count_nonzero(mem == -1)
            for e2 in range(e1 + 1, n):
                a = count_nonzero(mem + positions[e2] == -2)
                b = count_nonzero(mem == positions[e2])
                c = count_nonzero(positions[e2] == -1)
                e = count_nonzero(mem < positions[e2])
                relative_positions = array([e - d + a, m - e - b - c + a, b - a, c - a, d - a, a])
                put_before = vdot(relative_positions, cost_before)
                put_after = vdot(relative_positions, cost_after)
                put_tied = vdot(relative_positions, cost_tied)
                if put_before > put_after or put_before > put_tied:
                    edges.append((e2, e1))
                if put_after > put_before or put_after > put_tied:
                    edges.append((e1, e2))
                if put_tied < put_after and put_tied < put_before:
                    ties_must_be_checked = True
                matrix[e1][e2] = [put_before, put_after, put_tied]
                matrix[e2][e1] = [put_after, put_before, put_tied]
        graph_of_elements.add_edges(edges)
        return graph_of_elements, matrix, ties_must_be_checked

    @staticmethod
    def __positions(rankings: List[List[List or Set[int or str]]], elements_id: Dict) -> ndarray:
        positions = zeros((len(elements_id), len(rankings)), dtype=int) - 1
        id_ranking = 0
        for ranking in rankings:
            id_bucket = 0
            for bucket in ranking:
                for element in bucket:
                    positions[elements_id.get(element)][id_ranking] = id_bucket
                id_bucket += 1
            id_ranking += 1
        return positions

    def get_full_name(self) -> str:
        return "Exact algorithm ILP pulp"

    def is_scoring_scheme_relevant(self, scoring_scheme: ScoringScheme) -> bool:
        return True



from typing import List, Set, Dict
from corankco.kemeny_computation import KemenyScoreFactory
from corankco.dataset import Dataset
from corankco.scoringscheme import ScoringScheme
from enum import Enum, unique


@unique
class ConsensusFeature(Enum):
    AssociatedDataset = "dataset:"
    AssociatedScoringScheme = "scoring scheme:"
    AssociatedAlgorithm = "computed by:"
    IsNecessarilyOptimal = "necessarily optimal:"
    KemenyScore = "kemeny score:"
    WeakPartitioning = "weak partitioning (one optimal solution)"
    StrongPartitioning = "strong partitioning (all optimal solution)"


class Consensus:
    def __init__(self, consensus_rankings: List[List[List or Set[int or str]]],
                 dataset: Dataset = None,
                 scoring_scheme: ScoringScheme = None,
                 att: Dict[ConsensusFeature, str or int or float or bool] = None):
        self.__att = att
        if self.__att is None:
            self.__att = {}

        self.__set_consensus(consensus_rankings)

        self.__dataset = dataset
        self.__scoring_scheme = scoring_scheme

        if ConsensusFeature.KemenyScore not in self.__att:
            self.__att[ConsensusFeature.KemenyScore] = -1.
        if ConsensusFeature.IsNecessarilyOptimal not in self.__att:
            self.__att[ConsensusFeature.IsNecessarilyOptimal] = False

    def __get_consensus(self) -> List[List[List or Set[int or str]]]:
        return self.__consensus

    def __get_necessarily_optimal(self) -> bool:
        return self.__att[ConsensusFeature.IsNecessarilyOptimal]

    def __get_nb_consensus(self) -> int:
        return len(self.consensus_rankings)

    def __get_associated_score(self) -> float:
        if self.__att[ConsensusFeature.KemenyScore] == -1 and \
                self.__dataset is not None and \
                self.__scoring_scheme is not None:
            self.score = KemenyScoreFactory.get_kemeny_score(self.__scoring_scheme,
                                                             self.consensus_rankings[0],
                                                             self.__dataset)
        return self.__att[ConsensusFeature.KemenyScore]

    def __get_associated_dataset(self) -> Dataset:
        return self.__dataset

    def __get_associated_scoring_scheme(self) -> ScoringScheme:
        return self.__scoring_scheme

    def __get_att(self) -> Dict[ConsensusFeature, str or int or float or bool]:
        return self.__att

    def __set_necessarily_optimal(self, optimal: bool):
        self.__att[ConsensusFeature.IsNecessarilyOptimal] = optimal

    def __set_consensus(self, r: List[List[List or Set[int or str]]]):
        self.__consensus = r
        self.__set_nb_consensus(len(r))

    def __set_nb_consensus(self, nb_consensus: int):
        self.__nb_consensus = nb_consensus

    def __set_score(self, score: float):
        self.__att[ConsensusFeature.KemenyScore] = score

    consensus_rankings = property(__get_consensus, __set_consensus)
    nb_consensus = property(__get_nb_consensus, __set_nb_consensus)
    necessarily_optimal = property(__get_necessarily_optimal, __set_necessarily_optimal)
    score = property(__get_associated_score, __set_score)
    associated_dataset = property(__get_associated_dataset)
    associated_scoring_scheme = property(__get_associated_scoring_scheme)
    att = property(__get_att)

    def __str__(self)->str:
        return str(self.consensus_rankings)

    def description(self) -> str:
        self.__get_associated_score()
        return "Consensus description:" + "".join("\n\t"+str(key.value)+str(self.att[key]) for key in self.att) \
               + "\n\tconsensus:" + "".join("\n\t\tc"+str(i+1)+" = "
                                                     + str(self.consensus_rankings[i])
                                            for i in range(len(self.consensus_rankings)))



from enum import Enum, unique
from corankco.algorithms.allTied.alltied import AllTied
from corankco.algorithms.bioconsert.bioconsert import BioConsert
from corankco.algorithms.parcons.parcons import ParCons
from corankco.algorithms.exact.exactalgorithm import ExactAlgorithm


class AlgorithmEnumeration:
    median_ranking_algorithms = [
        AllTied,
        BioConsert,
        ParCons,
        ExactAlgorithm,
    ]


@unique
class Algorithm(Enum):
    AllTied = 0
    BioConsert = 1
    ParCons = 2
    Exact = 3



from corankco.algorithms.median_ranking import MedianRanking
from corankco.dataset import Dataset
from corankco.scoringscheme import ScoringScheme
from corankco.consensus import Consensus
from corankco.algorithms.exact.exactalgorithmgeneric import ExactAlgorithmGeneric


class ExactAlgorithm(MedianRanking):
    def __init__(self, limit_time_sec=0, scoring_scheme=None):
        if limit_time_sec > 0:
            self.__limit_time_sec = limit_time_sec
        else:
            self.__limit_time_sec = 0
        self.__scoring_scheme = scoring_scheme

    def compute_consensus_rankings(
            self,
            dataset: Dataset,
            scoring_scheme: ScoringScheme,
            return_at_most_one_ranking: bool=False,
            bench_mode: bool = False
    ) -> Consensus:
        """
        :param dataset: A dataset containing the rankings to aggregate
        :type dataset: Dataset (class Dataset in package 'datasets')
        :param scoring_scheme: The penalty vectors to consider
        :type scoring_scheme: ScoringScheme (class ScoringScheme in package 'distances')
        :param return_at_most_one_ranking: the algorithm should not return more than one ranking
        :type return_at_most_one_ranking: bool
        :param bench_mode: is bench mode activated. If False, the algorithm may return more information
        :type bench_mode: bool
        :return one or more rankings if the underlying algorithm can find several equivalent consensus rankings
        If the algorithm is not able to provide multiple consensus, or if return_at_most_one_ranking is True then, it
        should return a list made of the only / the first consensus found.
        In all scenario, the algorithm returns a list of consensus rankings
        :raise ScoringSchemeNotHandledException when the algorithm cannot compute the consensus because the
        implementation of the algorithm does not fit with the scoring scheme
        """

        try:
            import cplex
            from corankco.algorithms.exact.exactalgorithmcplex import ExactAlgorithmCplex
            return ExactAlgorithmCplex().compute_consensus_rankings(dataset,
                                                                    scoring_scheme,
                                                                    return_at_most_one_ranking,
                                                                    bench_mode)
        except:
            return ExactAlgorithmGeneric().compute_consensus_rankings(dataset,
                                                                      scoring_scheme,
                                                                      return_at_most_one_ranking,
                                                                      bench_mode)

    def get_full_name(self) -> str:
        return "Exact algorithm"

    def is_scoring_scheme_relevant(self, scoring_scheme: ScoringScheme) -> bool:
        return True



from corankco.dataset import Dataset
from corankco.scoringscheme import ScoringScheme
from corankco.consensus import Consensus


class ScoringSchemeNotHandledException(Exception):
    pass


class MedianRanking:
    def compute_consensus_rankings(
            self,
            dataset: Dataset,
            scoring_scheme: ScoringScheme,
            return_at_most_one_ranking: bool = False,
            bench_mode: bool = False
    ) -> Consensus:
        """
        :param dataset: A dataset containing the rankings to aggregate
        :type dataset: Dataset (class Dataset in package 'datasets')
        :param scoring_scheme: The penalty vectors to consider
        :type scoring_scheme: ScoringScheme (class ScoringScheme in package 'distances')
        :param return_at_most_one_ranking: the algorithm should not return more than one ranking
        :type return_at_most_one_ranking: bool
        :param bench_mode: is bench mode activated. If False, the algorithm may return more information
        :type bench_mode: bool
        :return Consensus containing a list of rankings. If the algorithm is not able to provide multiple consensus,
        or if return_at_most_one_ranking is True, a Consensus containing a single consensus ranking is returned.
        :raise ScoringSchemeNotHandledException when the algorithm cannot compute the consensus because the
        implementation of the algorithm does not fit with the scoring scheme
        """

        raise NotImplementedError("The method not implemented")

    def get_full_name(self) -> str:
        raise NotImplementedError("The method not implemented")

    def is_scoring_scheme_relevant(self, scoring_scheme: ScoringScheme) -> bool:
        """
        :return: a list of distances from distance_enumeration
        """
        raise NotImplementedError("The method not implemented")



class InvalidScoringScheme(Exception):
    pass


class NotRelevantScoringScheme(Exception):
    pass


class ScoringScheme:

    def __init__(self, penalties=None):
        if penalties is None:
            self._penalties = [[0., 1., 1., 0., 1., 0.], [1., 1., 0., 1., 1., 0.]]
        else:
            if type(penalties) is not list or len(penalties) != 2 or len(penalties[0]) != 6 or len(penalties[1]) != 6:
                raise InvalidScoringScheme
            if penalties[1][0] != penalties[1][1] or penalties[1][3] != penalties[1][4]:
                raise NotRelevantScoringScheme
            self._penalties = penalties

    @property
    def penalty_vectors_str(self):
        return self._penalties

    def __str__(self):
        return str(self._penalties)

    def description(self):
        # return str(self._penalties)
        return "\nScoring scheme description\n\tx before y in consensus\n\t\tx before y in input ranking: "\
               + str(self._penalties[0][0])\
               + "\n\t\ty before x in input ranking: "+str(self._penalties[0][1])\
               + "\n\t\tx and y tied in input ranking: " + str(self._penalties[0][2])\
               + "\n\t\tx present y missing in input ranking: " + str(self._penalties[0][3])\
               + "\n\t\tx missing y present ranking: " + str(self._penalties[0][4]) \
               + "\n\t\tx and y missing in input ranking: " + str(self._penalties[0][4]) \
               + "\n\tx and y tied in consensus\n\t\tx before y in input ranking: " + str(self._penalties[1][0]) \
                + "\n\t\ty before x in input ranking: " + str(self._penalties[1][1]) \
                + "\n\t\tx and y tied in input ranking: " + str(self._penalties[1][2]) \
                + "\n\t\tx present y missing in input ranking: " + str(self._penalties[1][3]) \
                + "\n\t\tx missing y present ranking: " + str(self._penalties[1][4]) \
                + "\n\t\tx and y missing in input ranking: " + str(self._penalties[1][4])

    @staticmethod
    def get_pseudodistance_scoring_scheme():
        return ScoringScheme(penalties=[[0., 1., 1., 0., 1., 0.], [1., 1., 0., 1., 1., 0.]])

    @staticmethod
    def get_unifying_scoring_scheme():
        return ScoringScheme(penalties=[[0., 1., 1., 0., 1., 1.], [1., 1., 0., 1., 1., 0.]])

    @staticmethod
    def get_induced_measure_scoring_scheme():
        return ScoringScheme(penalties=[[0., 1., 1., 0., 0., 0.], [1., 1., 0., 0., 0., 0.]])



from typing import Dict, List, Set, Tuple
from collections import deque
from corankco.scoringscheme import ScoringScheme
from corankco.dataset import Dataset
from numpy import zeros, vdot, ndarray, sort, count_nonzero, asarray


class DistanceRankingsError(Exception):
    pass


class KemenyScoreFactory:

    @staticmethod
    def get_kemeny_score(scoring_scheme: ScoringScheme, consensus: List[List or Set[int or str]], dataset: Dataset) \
            -> float:
        elements_r1 = {}
        size_buckets = {}
        id_bucket = 1
        sc = asarray(scoring_scheme.penalty_vectors_str)
        for bucket in consensus:
            size_buckets[id_bucket] = len(bucket)
            for element in bucket:
                elements_r1[element] = id_bucket
            id_bucket += 1
        before = zeros(6, dtype=int)
        tied = zeros(6, dtype=int)
        for ranking in dataset.rankings:
            tple = KemenyScoreFactory.__get_before_tied_counting(elements_r1, size_buckets, ranking, id_bucket)
            for i in range(6):
                before[i] += tple[0][i]
                tied[i] += tple[1][i]
        return abs(vdot(before, sc[0])) + abs(vdot(tied, sc[1]))

    @staticmethod
    def __get_before_tied_counting(r1: Dict, size_buckets: Dict, ranking: List[List or Set[int or str]], id_max: int) \
            -> Tuple[ndarray, ndarray]:
        vect_before = zeros(6, dtype=int)
        vect_tied = zeros(6, dtype=int)
        not_in_r2 = {}
        in_r1_only = set(r1.keys())
        n1 = len(r1)
        n2 = 0
        elem_r1_and_not_r2 = n1
        count_r2 = id_max
        ranking2 = {}
        present_in_both = 0
        id_ranking = 1
        for bucket in ranking:
            bucket_r2 = deque()
            n2 += len(bucket)
            for element in bucket:
                if element in r1:
                    in_r1_only.remove(element)
                    bucket_r2.appendleft(r1.get(element))
                    elem_r1_and_not_r2 -= 1
                    present_in_both += 1
                else:
                    bucket_r2.append(count_r2)
                    count_r2 += 1

            ranking2[id_ranking] = bucket_r2
            id_ranking += 1
        presence = zeros(count_r2, dtype=int)
        cumulated_up = zeros(count_r2, dtype=int)
        cumulated_down = zeros(count_r2, dtype=int)

        for bucket in ranking2.values():
            for element in bucket:
                if element < id_max:
                    presence[element] += 1
        cumulated_up[0] = presence[0]
        cumulated_down[0] = n2
        for i in range(1, count_r2):
            cumulated_up[i] = cumulated_up[i - 1] + presence[i]
            cumulated_down[i] = cumulated_down[i - 1] - presence[i]
        for element in in_r1_only:
            id_bucket = r1.get(element)
            if id_bucket not in not_in_r2:
                not_in_r2[id_bucket] = 1
            else:
                not_in_r2[id_bucket] += 1
            vect_before[3] += cumulated_up[id_bucket - 1]
            vect_before[4] += cumulated_down[id_bucket]

        tot_missing_r2 = elem_r1_and_not_r2
        for id_ties, size_ties_r1_both_missing_in_r2 in not_in_r2.items():
            tot_missing_r2 -= size_ties_r1_both_missing_in_r2
            vect_tied[5] += size_ties_r1_both_missing_in_r2 * (size_ties_r1_both_missing_in_r2 - 1) / 2
            vect_before[5] += size_ties_r1_both_missing_in_r2 * tot_missing_r2
            vect_tied[3] += (size_buckets[id_ties] - size_ties_r1_both_missing_in_r2)*size_ties_r1_both_missing_in_r2

        KemenyScoreFactory.__inversions(ranking2, 1, len(ranking2), vect_before, vect_tied, id_max)
        return vect_before, vect_tied

    @staticmethod
    def __inversions(ranking: Dict, left: int, right: int, vect1: ndarray, vect2: ndarray, id_max: int):
        if right == left:
            return KemenyScoreFactory.__manage_bucket(ranking.get(right), vect1, vect2, id_max)
        else:
            middle = (right - left) // 2
            return KemenyScoreFactory.__merge(
                KemenyScoreFactory.__inversions(ranking, left, middle + left, vect1, vect2, id_max),
                KemenyScoreFactory.__inversions(ranking, middle + left + 1, right, vect1, vect2, id_max),
                vect1,
                vect2,
                id_max)

    @staticmethod
    def __merge(left: ndarray, right: ndarray, vect_before: ndarray, vect_tied: ndarray, id_max: int):
        left_copy = left.copy()
        right_copy = right.copy()
        res = zeros(len(left_copy) + len(right_copy), dtype=int)
        n = len(left)
        m = len(right)
        i = 0
        j = 0
        k = 0
        not_in_r1_left = count_nonzero(left >= id_max)
        not_in_r1_right = count_nonzero(right >= id_max)
        vect_before[5] += not_in_r1_left * not_in_r1_right
        while i < n and j < m:
            nb = left[i]
            nb2 = right[j]
            if nb < nb2:
                if nb < id_max:
                    vect_before[0] += m - j - not_in_r1_right
                    vect_before[3] += not_in_r1_right
                res[k] = nb
                k += 1
                i += 1

            elif nb > nb2:
                if nb2 < id_max:
                    vect_before[1] += n - i - not_in_r1_left
                    vect_before[4] += not_in_r1_left
                res[k] = nb2
                k += 1
                j += 1
            else:
                cpt1 = 0
                cpt2 = 0
                while i < n and left[i] == nb:
                    res[k] = nb
                    k += 1
                    i += 1
                    cpt1 += 1
                while j < m and right[j] == nb:
                    res[k] = nb
                    k += 1
                    j += 1
                    cpt2 += 1
                if nb < id_max:
                    vect_tied[0] += cpt1 * cpt2
                    vect_before[0] += cpt1*(m - j - not_in_r1_right)
                    vect_before[1] += cpt2*(n - i - not_in_r1_left)
                    vect_before[3] += cpt1 * not_in_r1_right
                    vect_before[4] += cpt2 * not_in_r1_left

        while i < n:
            res[k] = left[i]
            k += 1
            i += 1
        while j < m:
            res[k] = right[j]
            k += 1
            j += 1
        return res

    @staticmethod
    def __manage_bucket(bucket: List or Set[int or str],
                        vect_before: ndarray,
                        vect_tied: ndarray,
                        id_max: int) -> ndarray:
        h = {}
        n = 0
        not_in_r1 = 0
        for elem in bucket:
            if elem < id_max:
                n += 1
                if elem not in h:
                    h[elem] = 1
                else:
                    h[elem] += 1
            else:
                not_in_r1 += 1
        vect_tied[5] += not_in_r1 * (not_in_r1 - 1) / 2
        vect_tied[3] += not_in_r1 * len(h)
        total = n
        for length_bucket_r1 in h.values():
            total -= length_bucket_r1

            vect_before[2] += length_bucket_r1 * total
            vect_tied[2] += length_bucket_r1 * (length_bucket_r1 - 1) / 2
        return sort(asarray(list(bucket)), kind='mergesort')

    @staticmethod
    def score_between_rankings(r1: List[List or Set[int or str]], r2: List[List or Set[int or str]], sc: ScoringScheme) \
            -> float:
        r1_complete = True
        r2_complete = True
        elem_r1 = set()
        elem_r2 = set()
        for bucket in r1:
            for elem in bucket:
                elem_r1.add(elem)
        for bucket in r2:
            for elem in bucket:
                elem_r2.add(elem)
                if elem not in elem_r1:
                    r1_complete = False

        for elem1 in elem_r1:
            if elem1 not in elem_r2:
                r2_complete = False
                break
        if r1_complete is True:
            return KemenyScoreFactory.get_kemeny_score(sc, r1, Dataset([r2]))
        elif r2_complete is True:
            return KemenyScoreFactory.get_kemeny_score(sc, r2, Dataset([r1]))
        else:
            raise DistanceRankingsError("The domain of one ranking must be included in the domain of the other one")






from setuptools import setup, find_packages


def readme():
    with open('README.rst') as f:
        return f.read()


setup(name='corankco',
      version='0.0.8',
      description='Kemeny-Young method for rank aggregation of incomplete rankings with ties',
      long_description=readme(),
      url='https://github.com/pierreandrieu/corankco',
      author='Pierre Andrieu',
      author_email='pierre.andrieu@lilo.org',
      license='MIT',
      packages=find_packages(include=['corankco', 'corankco.*']),
      python_requires='>=3',
      zip_safe=False,
      install_requires=['numpy',
                        'python-igraph',
                        'pulp==2.3',
                        'bioconsertinc==0.0.4',
                        ]
      )
